grammar Boppi;
import BoppiTokens;

program: expr EOF;

expr
	: singleExpr (COMPOUND singleExpr?)*
	;

singleExpr
	: PAROPEN expr PARCLOSE #parens
	| BRAOPEN expr BRACLOSE #block
	| IN PAROPEN variable (LISTDELIM variable)* PARCLOSE #read
	| OUT PAROPEN expr (LISTDELIM expr)* PARCLOSE #write
	| IFOPEN cond=expr IFTRUE onTrue=expr (IFFALSE onFalse=expr)? IFCLOSE #if
	| WHILEOPEN cond=expr WHILETRUE onTrue=expr WHILECLOSE #while
	| op=(PLUS|MINUS|NOT) singleExpr #prefix1
	| lhs=singleExpr op=(MULTIPLY|DIVIDE) rhs=singleExpr #infix1
	| lhs=singleExpr op=(PLUS|MINUS) rhs=singleExpr #infix2
	| lhs=singleExpr op=(LT|LEQ|GTE|GT|EQ|NEQ) rhs=singleExpr #infix3
	| lhs=singleExpr AND rhs=singleExpr #infix4
	| lhs=singleExpr OR rhs=singleExpr #infix5
	| DECLARE type IDENTIFIER #declare
	| <assoc=right> variable ASSIGN singleExpr #assign
	| variable #getVariable
	| LITERAL10 #literalInteger
	| CHAR #literalCharacter
	| (TRUE|FALSE) #literalBoolean
	| FUNCTION result=type name=IDENTIFIER PAROPEN (type IDENTIFIER (LISTDELIM type IDENTIFIER)*)? PARCLOSE body=singleExpr #declareFunction
	| variable PAROPEN (expr (LISTDELIM expr)*)? PARCLOSE #call
	;

type
	: ARROPEN LITERAL10 ARRCLOSE type #typeArray
	| type ARROW type #typeFunction
	| staticType=(INTTYPE | BOOLTYPE | CHARTYPE) #typeSimple
	| PAROPEN (type (LISTDELIM type)*)? PARCLOSE #typeTuple
	| variable #typeVariable
	;

variable
	: variable ARROPEN expr ARRCLOSE #variableArray
	| variable PROP IDENTIFIER #variableProperty
	| IDENTIFIER #variableSimple
	;
